@@ -743,15 +743,29 @@ static const char *opcodes[] = {
    "rll    r1,r3,d20(b2)",
    "rllg   r1,r3,d20(b2)",
 
-   "rnsbg  r1,r2,i3:u8,i4:u8,i5:u8",  // gie FIXME  un/signed  i3/4/5 ? t-bit ? z-bit?
-   "rxsbg  r1,r2,i3:u8,i4:u8,i5:u8",  // gie FIXME  ditto
-   "rosbg  r1,r2,i3:u8,i4:u8,i5:u8",  // gie FIXME  ditto
+   // Rotate and .... opcodes require special handling
+   //
+   // For rosbg and friends
+   // - Bit #0 of i3 is the T-bit and bit #1 of i3 ought to be 0.
+   // - i5 is optional and will not be written when 0
+   //
+   // For risbg and friends
+   // - Bit #0 of i4 is the Z-bit and bit #1 of i4 ought to be 0.
+   // - i5 is optional and will not be written when 0
+   //
+   // This implies that we need to model i3, i4 and i5 as masks so
+   // we can manipulate their value when disassembling or suppress
+   // the mask altogether. Note that we limit the set of allowed values
+   // for those masks to avoid excessively large numbers of testcases.
+   "rnsbg  r1,r2,m3:u8{0,1,2,63,128,129,191},m4:u6{0,1,2,63},m5:u6{0,1,2,63}",  // gie
+   "rxsbg  r1,r2,m3:u8{0,1,2,63,128,129,191},m4:u6{0,1,2,63},m5:u6{0,1,2,63}",  // gie
+   "rosbg  r1,r2,m3:u8{0,1,2,63,128,129,191},m4:u6{0,1,2,63},m5:u6{0,1,2,63}",  // gie
 
-   "risbg  r1,r2,i3:u8,i4:u8,i5:u8",  // gie FIXME  ditto
-   "risbgn r1,r2,i3:u8,i4:u8,i5:u8",  // mi1 FIXME  ditto
+   "risbg  r1,r2,m3:u6{0,1,2,63},m4:u8{0,1,2,63,128,129,191},m5:u6{0,1,2,63}",  // gie
+   "risbgn r1,r2,m3:u6{0,1,2,63},m4:u8{0,1,2,63,128,129,191},m5:u6{0,1,2,63}",  // mi1
 
-   "risbhg r1,r2,i3:u8,i4:u8,i5:u8",  // hiwo FIXME  ditto
-   "risblg r1,r2,i3:u8,i4:u8,i5:u8",  // hiwo FIXME  ditto
+   "risbhg r1,r2,m3:u5{0,1,2,31},m4:u8{0,1,2,31,128,129,159},m5:u6{0,1,2,63}",  // hiwo
+   "risblg r1,r2,m3:u5{0,1,2,31},m4:u8{0,1,2,31,128,129,159},m5:u6{0,1,2,63}",  // hiwo
 
    "srst   r1,r2",
 
@@ -951,6 +965,7 @@ static const char *opcodes[] = {
    "pfpo",                         // pfpo
    "srnm   d12(b2)",
    "srnmb  d12(b2)",               // fpext
+   "srnmt  d12(b2)",
    "sfpc   r1",
    // sfasr  not implemented
    "ste    f1,d12(x2,b2)",
@@ -962,6 +977,8 @@ static const char *opcodes[] = {
    // Chapter 10: Control Instructions                      not implemented
    // Chapter 14: I/O Instructions                          not implemented
    // Chapter 18: Hexadecimal-Floating-Point Instructions   not implemented
+   "lder    f1,f2",
+   "lde     f1,d12(x2,b2)",
 
    // Chapter 19: Binary-Floating-Point Instructions
    // Register pairs: 0-2 1-3 4-6 5-7 8-10 9-11 12-14 13-15
@@ -1140,12 +1157,12 @@ static const char *opcodes[] = {
    "dxtr    f1:{0,1,4,5,8,9,12,13},f2:{0,1,4,5,8,9,12,13},f3:{0,1,4,5,8,9,12,13}",
    "ddtra   f1,f2,f3,m4",
    "dxtra   f1:{0,1,4,5,8,9,12,13},f2:{0,1,4,5,8,9,12,13},f3:{0,1,4,5,8,9,12,13},m4",
-   "eedtr   f1,f2",
-   "eextr   f1,f2:{0,1,4,5,8,9,12,13}",
-   "esdtr   f1,f2",
-   "esxtr   f1,f2:{0,1,4,5,8,9,12,13}",
-   "iedtr   f1,f3,f2",
-   "iextr   f1:{0,1,4,5,8,9,12,13},f3:{0,1,4,5,8,9,12,13},f2",
+   "eedtr   r1,f2",
+   "eextr   r1,f2:{0,1,4,5,8,9,12,13}",
+   "esdtr   r1,f2",
+   "esxtr   r1,f2:{0,1,4,5,8,9,12,13}",
+   "iedtr   f1,f3,r2",
+   "iextr   f1:{0,1,4,5,8,9,12,13},f3:{0,1,4,5,8,9,12,13},r2",
    "ltdtr   f1,f2",
    "ltxtr   f1:{0,1,4,5,8,9,12,13},f2:{0,1,4,5,8,9,12,13}",
    // fidtr  not implemented
@@ -1160,8 +1177,8 @@ static const char *opcodes[] = {
    "mxtra   f1:{0,1,4,5,8,9,12,13},f2:{0,1,4,5,8,9,12,13},f3:{0,1,4,5,8,9,12,13},m4",
    "qadtr   f1,f3,f2,m4",
    "qaxtr   f1:{0,1,4,5,8,9,12,13},f3:{0,1,4,5,8,9,12,13},f2:{0,1,4,5,8,9,12,13},m4",
-   "rrdtr   f1,f3,f2,m4",
-   "rrxtr   f1:{0,1,4,5,8,9,12,13},f3:{0,1,4,5,8,9,12,13},f2,m4",
+   "rrdtr   f1,f3,r2,m4",
+   "rrxtr   f1:{0,1,4,5,8,9,12,13},f3:{0,1,4,5,8,9,12,13},r2,m4",
    "sldt    f1,f3,d12(x2,b2)",
    "slxt    f1:{0,1,4,5,8,9,12,13},f3:{0,1,4,5,8,9,12,13},d12(x2,b2)",
    "srdt    f1,f3,d12(x2,b2)",
@@ -1394,11 +1411,11 @@ parse_int(const char *p, int is_unsigned, long long *val)
       error("integer expected\n");
       return p;
    }
-   if (is_unsigned && val < 0) {
+   if (is_unsigned && *val < 0) {
       error("unsigned value expected\n");
       return p;
    }
-   return skip_digits(val < 0 ? p + 1 : p);
+   return skip_digits(*val < 0 ? p + 1 : p);
 }
 
 
@@ -1820,7 +1837,7 @@ get_opcode_by_name(const char *name)
 {
    unsigned len = strlen(name);
 
-   for (int i = 0; i < num_opcodes; ++i) {
+   for (unsigned i = 0; i < num_opcodes; ++i) {
       const char *op = opcodes[i];
       if (strncmp(op, name, len) == 0 &&
           (op[len] == ' ' || op[len] == '\0'))
@@ -1952,9 +1969,9 @@ parse_opcode(const char *spec)
 
    if (debug) {
       printf("opcode: |%s|\n", opc->name);
-      for (int i = 0; i < opc->num_opnds; ++i) {
+      for (unsigned i = 0; i < opc->num_opnds; ++i) {
          const opnd *d = opc->opnds + i;
-         printf("opnd %2d: %-8s  type: %-5s", i, d->name,
+         printf("opnd %2u: %-8s  type: %-5s", i, d->name,
                 opnd_kind_as_string(d->kind));
          if (d->kind != OPND_D12XB && d->kind != OPND_D12B &&
              d->kind != OPND_D20XB && d->kind != OPND_D20B &&
@@ -1963,7 +1980,7 @@ parse_opcode(const char *spec)
          if (d->allowed_values) {
             printf("  values:");
             unsigned nval = d->allowed_values[0];
-            for (int j = 1; j <= nval; ++j) {
+            for (unsigned j = 1; j <= nval; ++j) {
                if (d->is_unsigned)
                   printf(" %u", (unsigned)d->allowed_values[j]);
                else
@@ -1991,7 +2008,7 @@ get_opcode_by_index(unsigned ix)
 void
 release_opcode(opcode *opc)
 {
-   for (int i = 0; i < opc->num_opnds; ++i) {
+   for (unsigned i = 0; i < opc->num_opnds; ++i) {
       const opnd *q = opc->opnds + i;
       free(q->name);
       free(q->allowed_values);
@@ -2012,7 +2029,7 @@ static const char *unit_tests[] = {
    "err1 m", "err2 m2:", "err3 m3:s", "err4 m4:s5",
    "err5 m5{", "err6 m6:{", "err7 m7:{}", "err8 m8:{1", "err9 m9:{1,",
    "err10 m0:{2..}", "err11 m11:{2..1}", "err12 m11:u{1,2}",
-   "err13 m13:r", "err14 m14:u"
+   "err13 m13:r", "err14 m14:u", "err15 q1:u5{-1},q2:u8{3..1}"
 };
 
 
@@ -2023,6 +2040,6 @@ run_unit_tests(void)
 
    debug = 1;
 
-   for (int i = 0; i < num_tests; ++i)
+   for (unsigned i = 0; i < num_tests; ++i)
       parse_opcode(unit_tests[i]);
 }
